/* 
 * Rock-ulator (aka Vanguard Emulator) v.88a copyright 1997 by Brian Levine
 * with apologies to Moose for the new name
 *
 * The goal of this emulator is to faithfully reproduce the experience of playing
 * the Vanguard arcade machine, as well as the Rock-Ola games Fantasy and Nibbler, 
 * and to share the resulting code with whoever is interested in it.  No one is 
 * allowed to redistribute either the source code or the executable program for 
 * profit, nor is it to be distributed with Vanguard, Fantasy, or Nibbler ROM 
 * images.
 *
 * Brian Levine (blevine@systems.titan.com)
 *
 * 0.05 	Continuing the work...Game is now (barely) playable
 * 0.051	20% faster, quick start hack (yuck!), and setup for real colors
 * 0.06  Now using Multi-6502 Emulation Engine by Neil Bradley
 *       This fixes the game start problem!  Should be faster, too
 *       Score no longer displays as Hex.  Beginning origional color support
 * 0.65  Modified m6502.o so that no changes to header required.  Thanks, Allard!
 *       Joystick support added.  High score saved.  Screen dump added. All 
 *       colors are now correct.  Also, I finally adjusted version number to 
 *       reflect true status of the emulator.
 * 0.7   Custom blit function increases speed 25%.
 *  		Preliminary sound support added.  Preliminary gamepad support added.
 * 0.71  Fixed bug in custom blit and added swtich to disable it
 * 0.72  Sped up custom blit by another 20%.  Continued modifications to music.
 *       Fixed bug in screen dump.  Assorted other fixes.
 * 0.80  Major modifications to music (after discovering multi6502 bug).
 *       Background color enabled.  Various bug fixes.
 * 0.85  Added support for Fantasy and Nibbler.  
 * 0.86  Numerous fixes for multigame mode.  Game save feature.  Other bug fixes.
 * 0.88a More Fixes.  Color tables moved to source.  Throttle added
 * 0.89  Stole the tweaked VGA mode from M.A.M.E. 'cause its faster than ModeX.
 *       Fixed the sound in Fantasy and (I think) Nibbler.
 * 0.89a Added support for the 3rd Fantasy sound chip (sound effects), and
 *       really fixed the Nibbler sound.
 *
 * Vanguard Emulator 0.02 by Allard van der Bas (avdbas@wi.leidenuniv.nl)
 * Ported to GCC / Allegro.
 *
 * Original Vanguard Emulator 0.01 code (Borland C) by Ivan.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <conio.h>
#include <allegro.h>  
#include <audio.h>
#include <dos.h>
#include <go32.h>
#include <time.h>
#include <sys\movedata.h>
#include "m6502.h"
#include "twkuser.h"

//-- comment out for normal compile
//#define SOUNDTEST    
//#define DUMPDATA
//#define DUMPRW
//---------------------------------

unsigned short core_select;

//-- prototypes 
void xblast(BITMAP *source, int width, int x);
void xblit(BITMAP *source, BITMAP *dest, int source_x, int source_y, 
           int dest_x, int dest_y, int width, int height, int topline);
void vblit(BITMAP *source, BITMAP *dest, int source_x, int source_y, 
           int dest_x, int dest_y, int width, int height, int topline);
void xblitvsync(BITMAP *source, BITMAP *dest, int source_x, int source_y, 
           int dest_x, int dest_y, int width, int height, int topline);
void SetupVideo();
int  LoadRom(int Address, char * Name, size_t Size);
void DrawSprite (int c, int col, int row, BYTE color);
void DrawCharacter (int c, int col, int row, BYTE color);
void WarlordsDrawCharacter (int c, int col, int row, BYTE color);
void CCastlesDrawCharacter (int c, int col, int row);
void DisplayGraphics();
void WarlordsDisplayGraphics();
void CCastlesDisplayGraphics();
void DisplayForegroundChars(void);
void DisplayBackgroundChars(void);
void Exec();
int  SetupAudio();
void ShutdownAudio();
BYTE Interrupt(void);
int  VSoundOut0( WORD A,  BYTE B);
int  FSoundOut0( WORD A,  BYTE B);
int  NSoundOut0( WORD A,  BYTE B);
int  SoundOut1( WORD A,  BYTE B);
int  SoundOut2( WORD A,  BYTE B);
int  ColorOut( WORD A,  BYTE B);
int  SpeechOut( WORD A,  BYTE B);
WORD ForegroundUpdate( WORD A);
int  BackgroundUpdate( WORD A,  BYTE B);
WORD ReadCoin( WORD A);
WORD ReadPlayer1( WORD A);
WORD ReadPlayer2( WORD A);
void Help();
void LoadGame();
void SaveGame();
int  LoadHiScore();
void SaveHiScore();
WORD ReadScores( WORD A);
void PlaySample (int channel, BYTE *data, int len, int freq, int vol, int loop);
void AdjustSample (int channel, int freq, int vol);
void PlayAudio();
WORD GetBit(int startbit, unsigned int off);
void PrintVal(int add, int row, int col);
void PrintNum(int add, int row, int col);
void PrintStr(char *str, int row, int col);
void cBlit (int x, int y, int col, int row, int w, int h);
WORD ReadRom( WORD A);

void (*xb)(BITMAP *, BITMAP *, int, int, int, int, int, int, int);

#define INTCOUNT     12000

#define NUMVOICES    6
#define SAMPLE_RATE  44100
#define SAMPLE_LEN   50000

enum interrupt_types { INT_NONE=0, INT_IRQ, INT_NMI, INT_QUIT };

typedef struct
{
   int  loadAddr;
   char *fileName;
   int  romSize;
   char *desc;
} ROMS;

enum Games { VANGUARD=0, FANTASY, NIBBLER, WARLORDS, CCASTLES };

typedef struct m6502MemoryWrite MemoryWrite;
typedef struct m6502MemoryRead MemoryRead;

typedef struct 
{
   WORD GameID;
   ROMS *Roms;
   MemoryWrite *MemWrite;
   MemoryRead *MemRead;
   void (*Display)();
   BYTE *Color;
   WORD ForegroundRAM;
   WORD BackgroundRAM;
   WORD ColorRAM;
   int  BChar1;
   int  BChar2;
   WORD Sound1;
   WORD Sound2;
   WORD Sound3;
   WORD Port1;
   WORD Port2;
   WORD Port3;
   WORD Port4;
   WORD Player1;
   WORD Player2;
   WORD DipSwitch;
   WORD Coin;
   WORD VScroll;
   WORD HScroll;
   WORD Speech;
   WORD HiScoreTable;
   WORD HiScoreTableLength;
   WORD HiScoreTable0;
   WORD HiScore;
   WORD HiScoreLength;
   char *Path;
} GameAddresses;

Register scr256x256[] =
{
	{ 0x3c2, 0x00, 0xe3},{ 0x3d4, 0x00, 0x5f},{ 0x3d4, 0x01, 0x3f},
	{ 0x3d4, 0x02, 0x40},{ 0x3d4, 0x03, 0x82},{ 0x3d4, 0x04, 0x4A},
	{ 0x3d4, 0x05, 0x9A},{ 0x3d4, 0x06, 0x23},{ 0x3d4, 0x07, 0xb2},
	{ 0x3d4, 0x08, 0x00},{ 0x3d4, 0x09, 0x61},{ 0x3d4, 0x10, 0x0a},
	{ 0x3d4, 0x11, 0xac},{ 0x3d4, 0x12, 0xff},{ 0x3d4, 0x13, 0x20},
	{ 0x3d4, 0x14, 0x40},{ 0x3d4, 0x15, 0x07},{ 0x3d4, 0x16, 0x1a},
	{ 0x3d4, 0x17, 0xa3},{ 0x3c4, 0x01, 0x01},{ 0x3c4, 0x04, 0x0e},
	{ 0x3ce, 0x05, 0x40},{ 0x3ce, 0x06, 0x05},{ 0x3c0, 0x10, 0x41},
	{ 0x3c0, 0x13, 0x00}
};

BYTE vanguardColors[] = 
{
   0x00, 0x3F, 0x80, 0xC6, 0xEF, 0x2F, 0xC6, 0xF8, 
   0xFE, 0xE7, 0xC6, 0xC0, 0xFF, 0x38, 0x2F, 0xC6, 
   0x00, 0x80, 0x07, 0x2F, 0xEF, 0xF8, 0x07, 0xFF, 
   0xFE, 0xF8, 0xFF, 0xC0, 0xFF, 0xC6, 0xE7, 0xF4, 
   0x00, 0xF4, 0x2F, 0xFF, 0xEF, 0xFF, 0xF8, 0x07, 
   0xFE, 0x07, 0xC0, 0x3F, 0xFF, 0xC6, 0x3F, 0xC0, 
   0x00, 0xE7, 0x38, 0x07, 0xEF, 0xF4, 0xC0, 0xFF, 
   0xFE, 0xF8, 0xFF, 0xC0, 0xFF, 0xE7, 0xC6, 0xC0
};

BYTE fantasyColors[] = 
{
   0x00, 0x27, 0x06, 0x05, 0x7F, 0x38, 0x2F, 0x05, 
   0x00, 0x05, 0x38, 0x28, 0x00, 0xFF, 0x07, 0xF4, 
   0x00, 0x38, 0x07, 0x2F, 0x00, 0xFF, 0xF8, 0xC5, 
   0x00, 0x38, 0xC0, 0xF4, 0x00, 0xFF, 0x27, 0x2F, 
   0x00, 0x38, 0x07, 0xF8, 0x00, 0x2F, 0x05, 0xF8, 
   0x00, 0x2F, 0xC0, 0xF8, 0x00, 0x07, 0xF8, 0xFF, 
   0x00, 0xF4, 0x07, 0xC0, 0x00, 0xC5, 0x38, 0x07, 
   0x00, 0x2F, 0xFF, 0x3F, 0x00, 0xC5, 0xF8, 0xC0
};

BYTE nibblerColors[] =
{
   0x00, 0xFF, 0x3F, 0xC0, 0x00, 0x38, 0xC7, 0x05,
   0x00, 0xC0, 0x07, 0x3F, 0x00, 0xE0, 0x3F, 0x05,
   0x00, 0xAC, 0x07, 0xC0, 0x00, 0xC5, 0xFF, 0x2F,
   0x00, 0x05, 0xC0, 0x2F, 0x00, 0x05, 0x3F, 0xC7,
   0x00, 0xFF, 0x07, 0xC5, 0x00, 0xAD, 0x38, 0xA8,
   0x00, 0x3F, 0xAD, 0xC0, 0x00, 0x07, 0xFF, 0xFF,
   0x00, 0xC0, 0x3F, 0xAD, 0x00, 0xC5, 0xFF, 0x3F,
   0x00, 0x3F, 0xFF, 0x07, 0x00, 0xC5, 0x07, 0x3F
};

ROMS vanguardRoms[] = 
{
   { 0x2000, "sk5_ic50.bin", 2048, "background character graphics - plane 1" },
   { 0x2800, "sk5_ic51.bin", 2048, "background character graphics - plane 2" },
   { 0x4000, "sk4_ic07.bin", 4096, "code" },
   { 0x5000, "sk4_ic08.bin", 4096, "code" },
   { 0x6000, "sk4_ic09.bin", 4096, "code" },
   { 0x7000, "sk4_ic10.bin", 4096, "code, sprite graphics, tables" },
   { 0x8000, "sk4_ic13.bin", 4096, "foreground character graphics" },
   { 0x9000, "sk4_ic14.bin", 4096, "tables (sound?)" },
   { 0xA000, "sk4_ic15.bin", 4096, "more tables?" },
   { 0xB000, "sk4_ic16.bin", 4096, "data (1K), code" },
   { 0xC000, "sk4_ic51.bin", 2048, "sound - part 1" },
   { 0xC800, "sk4_ic52.bin", 2048, "sound - part 2" },
   { 0xF000, "sk4_ic13.bin", 4096, "vectors" },
   { 0, NULL, 0, NULL }	
};

ROMS fantasyRoms[] = 
{
   { 0x3000, "ic12.cpu",	4096, "code" },
   { 0x4000, "ic07.cpu", 	4096, "code" },
   { 0x5000, "ic08.cpu", 	4096, "code" },
   { 0x6000, "ic09.cpu", 	4096, "code" },
   { 0x7000, "ic10.cpu",  	4096, "code / graphics" },
   { 0x8000, "ic14.cpu",  	4096, "graphics" },
   { 0x9000, "ic15.cpu",  	4096, "data" },
   { 0xA000, "ic16.cpu",  	4096, "code - graphics" },
   { 0xB000, "ic17.cpu",  	4096, "graphics - code" },
   { 0xC000, "ic50.vid",   4096, "graphics" },
   { 0xD000, "ic51.vid",   4096, "graphics" },
   { 0xF000, "ic14.cpu",  	4096, "vectors" },
   { 0xE000, "ic52.cpu",   2048, "sound" },
   { 0xE800, "ic51.cpu",   2048, "sound" },
   { 0xF000, "ic53.cpu",   2048, "sound" },
   { 0, NULL, 0, NULL }	
};

ROMS nibblerRoms[] = 
{
   { 0x3000, "ic12",	   4096, "code" },
   { 0x4000, "ic07", 	4096, "code" },
   { 0x5000, "ic08", 	4096, "code" },
   { 0x6000, "ic09", 	4096, "code" },
   { 0x7000, "ic10",  	4096, "code / graphics" },
   { 0x8000, "ic14",  	4096, "graphics" },
   { 0x9000, "ic15",  	4096, "data" },
   { 0xa000, "ic16",  	4096, "code - graphics" },
   { 0xb000, "ic17",  	4096, "graphics - code" },
   { 0xC000, "ic50",    4096, "graphics" },
   { 0xD000, "ic51",    4096, "graphics" },
   { 0xF000, "ic14",  	4096, "vectors" },
   { 0xE000, "ic52",    2048, "sound" },
   { 0xE800, "ic53",    2048, "sound" },
   { 0, NULL, 0, NULL }	
};

ROMS warlordsRoms[] = 
{
   { 0x5000, "037154.1m", 2048, "code" },
   { 0x5800, "037153.1k", 2048, "code" },
   { 0x6000, "037158.1j", 2048, "code" },
   { 0x6800, "037157.1h", 2048, "code - tables" },
   { 0x7000, "037156.1e", 2048, "code" },
   { 0x7800, "037155.1d", 2048, "code" },
   { 0x8000, "037159.6e", 2048, "graphics" },
   { 0x8800, "037159.6h", 2048, "graphics" },
   { 0xF800, "037155.1d", 2048, "vectors" },
   { 0, NULL, 0, NULL }	
};

ROMS ccastlesRoms[] = 
{
   { 0x10000, "ccastles.101", 8192, "graphics" },
   { 0x12000, "ccastles.102",	8192, "graphics" },
   { 0x14000, "ccastles.106",	8192, "graphics" },
   { 0x16000, "ccastles.107",	8192, "graphics" },
   { 0xA000,  "ccastles.303",	8192, "code" },
   { 0xC000,  "ccastles.304",	8192, "code" },
   { 0xE000,  "ccastles.305",	8192, "code - vectors" },
   { 0, NULL, 0, NULL }	
};

MemoryWrite VanguardWrite[] = 
{
   { 0x3100, 0x3100, VSoundOut0 }, 
   { 0x3101, 0x3101, SoundOut1 }, 
   { 0x3102, 0x3102, SoundOut2 }, 
   { 0x3103, 0x3103, ColorOut }, 
   { 0x3400, 0x3400, SpeechOut }, 
   { -1, -1, NULL }
};

MemoryRead VanguardRead[] = 
{
   { 0x3104, 0x3104, ReadPlayer1 },
   { 0x3105, 0x3105, ReadPlayer2 },
   { 0x3107, 0x3107, ReadCoin },
   { 0x0220, 0x0220, ReadScores },
   { -1, -1, NULL }
};

MemoryWrite FantasyWrite[] = 
{
   { 0x2100, 0x2100, FSoundOut0 }, 
   { 0x2101, 0x2101, SoundOut1 }, 
   { 0x2102, 0x2102, SoundOut2 }, 
   { 0x2103, 0x2103, ColorOut }, 
   { 0x2400, 0x2400, SpeechOut }, 
   { -1, -1, NULL }
};

MemoryRead FantasyRead[] = 
{
   { 0x2104, 0x2104, ReadPlayer1 },
   { 0x2105, 0x2105, ReadPlayer2 },
   { 0x2107, 0x2107, ReadCoin },
   { 0x0220, 0x0220, ReadScores },
   { -1, -1, NULL }
};

MemoryWrite NibblerWrite[] = 
{
   { 0x2100, 0x2100, NSoundOut0 }, 
   { 0x2101, 0x2101, SoundOut1 }, 
   { 0x2102, 0x2102, SoundOut2 }, 
   { 0x2103, 0x2103, ColorOut }, 
   { 0x2400, 0x2400, SpeechOut }, 
   { -1, -1, NULL }
};

MemoryRead NibblerRead[] = 
{
   { 0x2104, 0x2104, ReadPlayer1 },
   { 0x2105, 0x2105, ReadPlayer2 },
   { 0x2107, 0x2107, ReadCoin },
   { 0x0290, 0x02EF, ReadScores },
   { -1, -1, NULL }
};

MemoryWrite WarlordsWrite[] = 
{
   { 0x0000, 0x4FFF, BackgroundUpdate },
   { -1, -1, NULL }
};

MemoryRead WarlordsRead[] = 
{
   { 0x0000, 0x4FFF, ForegroundUpdate },
   { -1, -1, NULL }
};

MemoryWrite CCastlesWrite[] = 
{
   { 0x0000, 0x9FFF, BackgroundUpdate },
   { -1, -1, NULL }
};

MemoryRead CCastlesRead[] = 
{
   { 0x0000, 0x9FFF, ForegroundUpdate },
   { -1, -1, NULL }
};

GameAddresses gaVanguard = 
{
   VANGUARD, vanguardRoms,
   VanguardWrite, VanguardRead,
   DisplayGraphics,
   vanguardColors,
   0x0400, 0x0800, 0x0C00,
   0x2000, 0x2800,
   0xC000, 0xC800, 0x0000,
   0x3100, 0x3101, 0x3102, 0x3103,
   0x3104, 0x3105, 0x3106, 0x3107,
   0x3200, 0x3300, 0x3400,
   0x220, 0x70, 0x22B, 0x25, 3,
   "Vanguard"
};

GameAddresses gaFantasy = 
{
   FANTASY, fantasyRoms,
   FantasyWrite, FantasyRead,
   DisplayGraphics,
   fantasyColors,
   0x0400, 0x0800, 0x0C00,
   0xC800, 0xD800,
   0xE000, 0xE800, 0xF000,
   0x2100, 0x2101, 0x2102, 0x2103,
   0x2104, 0x2105, 0x2106, 0x2107,
   0x2200, 0x2300, 0x2400,
   0x220, 0x70, 0x22B, 0x25, 3,
   "Fantasy"
};

GameAddresses gaNibbler = 
{
   NIBBLER, nibblerRoms,
   NibblerWrite, NibblerRead,
   DisplayGraphics,
   nibblerColors,
   0x0400, 0x0800, 0x0C00,
   0xC800, 0xD800,
   0x0000, 0xE000, 0x0000,
   0x2100, 0x2101, 0x2102, 0x2103,
   0x2104, 0x2105, 0x2106, 0x2107,
   0x2200, 0x2300, 0x2400,
   0x290, 0x60, 0x290, 0x2B4, 4,
   "Nibbler"
};

GameAddresses gaWarlords = 
{
   WARLORDS, warlordsRoms,
   WarlordsWrite, WarlordsRead,
   WarlordsDisplayGraphics,
   vanguardColors,
   0x0000, 0x0400, 0x0400,
   0x8000, 0x8800,
   0xC000, 0xC800, 0x0000,
   0x3100, 0x3101, 0x3102, 0x3103,
   0x3104, 0x3105, 0x3106, 0x0C00,
   0x3200, 0x3300, 0x3400,
   0x220, 0x70, 0x22B, 0x25, 3,
   "Warlords"
};

GameAddresses gaCCastles = 
{
   CCASTLES, ccastlesRoms,
   CCastlesWrite, CCastlesRead,
   CCastlesDisplayGraphics,
   vanguardColors,
   0x0000, 0x8000, 0x0000,
   0x14000, 0x16000,
   0xC000, 0xC800, 0x0000,
   0x3100, 0x3101, 0x3102, 0x3103,
   0x3104, 0x3105, 0x3106, 0x9600,
   0x3200, 0x3300, 0x3400,
   0x220, 0x70, 0x22B, 0x25, 3,
   "CCastles"
};

// Global variables
GameAddresses *cga;
BITMAP *piccie;
BITMAP *fchars;
BITMAP *bchars;
BYTE RAM[0x20000];
BYTE Foreground[0x10000];
BYTE Background[0x10000];

PALETTE pal;
WORD ExitAdress;

int sw_dumpgraphics = 0;
int sw_gravis = 0;
int sw_hr = 0;
int sw_men = 0x00;
int sw_nojoystick = 0;
int sw_nosound = 0;
int sw_noxblit = 0;
int sw_sb16 = 0;
int sw_sb = 0;
int sw_skip = 0;
int sw_throttle = 0;
int sw_useX = 0;
int sw_v360 = 0;
int sw_vsync = 0;
int sw_xvga = 0;

int joy_next_shot = 0;
int score_loaded = 0;
int sound_test = 0;

int intr_ticks = 0;
int throttle = 0;
int fps = 0;
int ShowFPS = 0;

int NoSound = TRUE;
int NoSound0 = TRUE;
int Sound0;
int Sound0Base;
int oSound0Base;
int NoSound1 = TRUE;
int Sound1;
int Sound1Base;
int oSound1Base;
int NoSound2 = TRUE;
int Sound2;
int Sound2Base;
int oSound2Base;
HAC hVoice[NUMVOICES];
LPAUDIOWAVE lpWave[NUMVOICES];

int swapsound[8] = { 0, 2, 3, 6, 1, 4, 5, 7 };

int LoopCounter = 0;
BYTE LastPort1 = 0;
BYTE LastPort2 = 0;
BYTE LastPort3 = 0;
BYTE LastPort4 = 0;
FILE *fptr;

int ActivePage = 0;
BOOL pause = FALSE;

//-- sine wave sample
static BYTE samples[32] = 
{
   0x00, 0x00, 0x00, 0x00, 0x22, 0x22, 0x22, 0x22, 
   0x44, 0x44, 0x44, 0x44, 0x22, 0x22, 0x22, 0x22, 
   0x00, 0x00, 0x00, 0x00, 0xdd, 0xdd, 0xdd, 0xdd, 
   0xbb, 0xbb, 0xbb, 0xbb, 0xdd, 0xdd, 0xdd, 0xdd
};

//-- color intensity based on 2 or 3 bit definition in color ROMs
BYTE intensity4[] = { 0, 31, 47, 63 };
BYTE intensity8[] = { 0, 15, 23, 31, 39, 47, 55, 63 };

void SetupVideo () 
{
   WORD i;
   BYTE b=0;	
	Register *reg = 0;
	int reglen = 0;
   int videofreq = 0;

   if (sw_v360)
      set_gfx_mode(GFX_MODEX, 360,360, 0,720);
   else if (sw_xvga)
      {
		set_gfx_mode(GFX_VGA,320,200,0,0);
		reg = scr256x256;
		reglen = sizeof(scr256x256)/sizeof(Register);
		reg[0].value = (reg[0].value & 0xf3) | (videofreq << 2);
		outRegArray(reg,reglen);
      }
   else
      set_gfx_mode(GFX_MODEX, 256,256, 0,512);

   for (i=0; i<256; i++, b++)
      {
      pal[i].r = intensity8[b&0x7];
      pal[i].g = intensity8[(b&0x38)>>3];
      pal[i].b = intensity4[(b&0xc0)>>6];
      }

   set_palette(pal);

   piccie = create_bitmap(256,256);
   fchars = create_bitmap(256, 64);
   bchars = create_bitmap(256, 64);

   clear (piccie);
   clear (fchars);
}

void SaveGame()
{
   char file[80];
   FILE *fp;
   sprintf (file, "%s\\savegame", cga->Path);
   fp = fopen (file, "wb");
   fwrite (RAM, 0x4000, 1, fp);
	fwrite (&m6502clockticks, 4, 1, fp);
	fwrite (&m6502pc, 4, 1, fp);
	fwrite (&m6502af, 4, 1, fp);
	fwrite (&m6502x, 4, 1, fp);
	fwrite (&m6502y, 4, 1, fp);
	fwrite (&m6502s, 4, 1, fp);
	fwrite (&cyclesRemaining, 4, 1, fp);
	fwrite (&syncCycle, 4, 1, fp);
	fwrite (&NoSound0, 4, 1, fp);
	fwrite (&Sound0, 4, 1, fp);
	fwrite (&Sound0Base, 4, 1, fp);
	fwrite (&oSound0Base, 4, 1, fp);
	fwrite (&NoSound1, 4, 1, fp);
	fwrite (&Sound1, 4, 1, fp);
	fwrite (&Sound1Base, 4, 1, fp);
	fwrite (&oSound1Base, 4, 1, fp);
	fwrite (&NoSound2, 4, 1, fp);
	fwrite (&Sound2, 4, 1, fp);
	fwrite (&Sound2Base, 4, 1, fp);
	fwrite (&oSound2Base, 4, 1, fp);
   fclose (fp);
}

void LoadGame()
{
   char file[80];
   FILE *fp;
   sprintf (file, "%s\\savegame", cga->Path);
   fp = fopen (file, "rb");
   if (fp != NULL)
      {
      fread (RAM, 0x4000, 1, fp);
   	fread (&m6502clockticks, 4, 1, fp);
	   fread (&m6502pc, 4, 1, fp);
   	fread (&m6502af, 4, 1, fp);
	   fread (&m6502x, 4, 1, fp);
   	fread (&m6502y, 4, 1, fp);
	   fread (&m6502s, 4, 1, fp);
   	fread (&cyclesRemaining, 4, 1, fp);
	   fread (&syncCycle, 4, 1, fp);
	   fread (&NoSound0, 4, 1, fp);
	   fread (&Sound0, 4, 1, fp);
   	fread (&Sound0Base, 4, 1, fp);
	   fread (&oSound0Base, 4, 1, fp);
   	fread (&NoSound1, 4, 1, fp);
	   fread (&Sound1, 4, 1, fp);
   	fread (&Sound1Base, 4, 1, fp);
	   fread (&oSound1Base, 4, 1, fp);
   	fread (&NoSound2, 4, 1, fp);
	   fread (&Sound2, 4, 1, fp);
   	fread (&Sound2Base, 4, 1, fp);
	   fread (&oSound2Base, 4, 1, fp);
      fclose (fp);
      }
}

void SaveHiScore()
{
   char file[80];
   FILE *fp;
   sprintf (file, "%s\\hiscore", cga->Path);
   fp = fopen (file, "wb");
   fwrite (RAM + cga->HiScoreTable, cga->HiScoreTableLength, 1, fp);
   fclose (fp);
}

WORD ReadScores(WORD A)
{
   if (!score_loaded)
      score_loaded = LoadHiScore();

   return (WORD)RAM[A];
}

int LoadHiScore()
{
   char file[80];
   FILE *fp;

   if (cga->GameID == NIBBLER)
       if (memcmp (RAM + cga->HiScoreTable0, "\x00\x50\x00\x00", cga->HiScoreLength) ||
           memcmp (RAM + cga->HiScore, "\x00\x05\x00\x00", cga->HiScoreLength))
           return 0;

   sprintf (file, "%s\\hiscore", cga->Path);
   fp = fopen (file, "rb");
   if (fp == NULL)
      return 1;
   fread (RAM + cga->HiScoreTable, cga->HiScoreTableLength, 1, fp);
   fclose (fp);

   //-- now copy high score from table to current high score 
   if (cga->GameID == NIBBLER)
      memcpy (RAM + cga->HiScore, RAM + cga->HiScoreTable0, cga->HiScoreLength);

   return 1;
}

void Help()
{
   clrscr();
	printf ("Rock-ulator by Brian Levine\n\n");
   printf ("   Multi-6502 Emulation Engine by Neil Bradley\n");
   printf ("   Allegro Library by Shawn Hargreaves\n");
   printf ("   SEAL Synthetic Audio Library by Carlos Hasen\n\n");
   printf ("Command Line Switches:\n");
   printf ("   -d         display ROM graphics before starting emulation\n");
   printf ("   -fantasy   play Fantasy\n");
   printf ("   -m3/4/5/6  3/4/5 ships per game (default 3)\n");
   printf ("   -nibbler   play Nibbler\n");
   printf ("   -nojoy     disable joystick support\n");
   printf ("   -nosound   disable sound support\n");
   printf ("   -noxblit   disable custom blit\n");
   printf ("   -sb/sb16   force sound support to Soundblaster/Soundblaster16\n");
   printf ("   -skip      skip every other frame\n");
   printf ("   -throttle  keep emulation from running faster than 60 fps\n");
   printf ("   -v360      use 360 x 360 video mode\n");
   printf ("   -vsync     forces custom blit to wait for vsync\n");
   printf ("   -x         use X key to fire down\n\n");
   printf ("<Hit any key to begin emulation>\n");
   getch();
}

int SetupAudio()
{
	int i;
   unsigned int ndevice;
   AUDIOINFO info;

   if (AInitialize())
      return 1;

	APingAudio(&ndevice);

   info.wFormat = AUDIO_FORMAT_8BITS | AUDIO_FORMAT_MONO;
   info.nSampleRate = SAMPLE_RATE;

   if (sw_sb16)
      {
      info.wFormat = AUDIO_FORMAT_16BITS | AUDIO_FORMAT_STEREO;
   	info.nDeviceId = AUDIO_PRODUCT_SB16;
      }
	else if (sw_sb)
   	info.nDeviceId = AUDIO_PRODUCT_SB;
   else
      info.nDeviceId = AUDIO_DEVICE_MAPPER;


   if (AOpenAudio(&info))
      return 1;

	AOpenVoices(NUMVOICES);
   //-- allocate voices and waveforms
   for(i=0; i<NUMVOICES; i++)
      {
      ACreateAudioVoice(&hVoice[i]);
      ASetVoicePanning(hVoice[i], 0x80);

      if ((lpWave[i] = (LPAUDIOWAVE)malloc(sizeof(AUDIOWAVE))) == 0)
         {
         //-- ERROR - free audio memory and exit
         int j;
         for(j=0; j<i; j++)
            {
            ADestroyAudioVoice(hVoice[j]);
            ADestroyAudioData(lpWave[j]);
            free(lpWave[j]);
            }
         ADestroyAudioVoice(hVoice[i]);
         return 1;
         }

      if (i < NUMVOICES-2 || i == 5)
         {
         lpWave[i]->wFormat = AUDIO_FORMAT_8BITS | AUDIO_FORMAT_MONO;
         lpWave[i]->nSampleRate = SAMPLE_RATE;
         lpWave[i]->dwLength = SAMPLE_LEN;
         lpWave[i]->dwLoopStart = lpWave[i]->dwLoopEnd = 0;

         if (ACreateAudioData(lpWave[i]) != AUDIO_ERROR_NONE)
            {
            //-- ERROR - free audio memory
            int j;
            for(j=0; j<i; j++)
               {
               ADestroyAudioVoice(hVoice[j]);
               ADestroyAudioData(lpWave[j]);
               free(lpWave[j]);
               }
            ADestroyAudioVoice(hVoice[i]);
            free(lpWave[i]);
            return 1;
            }
         }
      }

   ALoadWaveFile("explsion.wav", &(lpWave[3]), 0L);
   ALoadWaveFile("fire.wav", &(lpWave[4]), 0L);
   ALoadWaveFile("fire.wav", &(lpWave[2]), 0L);

   NoSound = FALSE;
   NoSound0 = TRUE;
   Sound0 = cga->Sound1;
   Sound0Base = cga->Sound1;
   oSound0Base = cga->Sound1;
   NoSound1 = TRUE;
   Sound1 = cga->Sound2;
   Sound1Base = cga->Sound2;
   oSound1Base = cga->Sound2;
   NoSound2 = TRUE;
   Sound2 = cga->Sound3;
   Sound2Base = cga->Sound3;
   oSound2Base = cga->Sound3;
   
   return 0;
}

void ShutdownAudio()
{
	int i;

   for(i=0; i<NUMVOICES; i++)
      {
      AStopVoice(hVoice[i]);
      ADestroyAudioVoice(hVoice[i]);
      ADestroyAudioData(lpWave[i]);
      free(lpWave[i]);
      }

   //-- stop audio
   ACloseVoices();
   ACloseAudio();
}

int htoi(char c)
{
   if (c > '9')
      {
      c &= 0xDF;         
      return (c - 'A' + 10);
      }
   return (c - '0');
}

int main(int argc, char **argv) 
{

   int i;

   core_select = (unsigned short)_dos_ds;

   if (argc < 2)
      Help();

   //-- check arguments
   cga = &gaVanguard;
   xb = xblit;

   for (i=1; i<argc; i++)
      {
      if (strcmp ("-fantasy", argv[i]) == 0)
         cga = &gaFantasy;
      else if (strcmp ("-nibbler", argv[i]) == 0)
         cga = &gaNibbler;
      else if (strcmp ("-warlords", argv[i]) == 0)
         cga = &gaWarlords;
      else if (strcmp ("-ccastles", argv[i]) == 0)
         cga = &gaCCastles;
      else if (memcmp ("-ds", argv[i], 3) == 0)
         sw_men = (htoi(argv[i][3])<<4)+htoi(argv[i][4]);
      else if (strcmp ("-m4", argv[i]) == 0)
         {
         switch (cga->GameID)
            {
            case VANGUARD:
               sw_men |= 0x10;
               break;
            case FANTASY:
               sw_men |= 0x20;
               break;
            case NIBBLER:
               sw_men |= 0x01;
               break;
            }
         }
      else if (strcmp ("-m5", argv[i]) == 0)
         {
         switch (cga->GameID)
            {
            case VANGUARD:
               sw_men |= 0x20;
               break;
            case FANTASY:
               sw_men |= 0x40;
               break;
            case NIBBLER:
               sw_men |= 0x02;
               break;
            }
         }
      else if (strcmp ("-m6", argv[i]) == 0)
         {
         switch (cga->GameID)
            {
            case VANGUARD:
               sw_men |= 0x20;
               break;
            case FANTASY:
               sw_men |= 0x40;
               break;
            case NIBBLER:
               sw_men |= 0x03;
               break;
            }
         }
      else if (strcmp ("-d", argv[i]) == 0)
         sw_dumpgraphics = 1;
      else if (strcmp ("-x", argv[i]) == 0)
         sw_useX = 1;
      else if (strcmp ("-throttle", argv[i]) == 0)
         sw_throttle = 1;
      else if (strcmp ("-skip", argv[i]) == 0)
         sw_skip = 1;
      else if (strcmp ("-nosound", argv[i]) == 0)
         sw_nosound = 1;
      else if (strcmp ("-sb16", argv[i]) == 0 && !sw_nosound)
         sw_sb16 = 1;
      else if (strcmp ("-sb", argv[i]) == 0 && !sw_nosound)
            sw_sb = 1;
      else if (strcmp ("-v360", argv[i]) == 0)
         sw_v360 = 1;
      else if (strcmp ("-vsync", argv[i]) == 0)
         {
         sw_vsync = 1;
         xb = xblitvsync;
         }
      else if (strcmp ("-xvga", argv[i]) == 0)
         sw_xvga = 1;
      else if (strcmp ("-noxblit", argv[i]) == 0)
         sw_noxblit = 1;
      else if (strcmp ("-nojoy", argv[i]) == 0)
         sw_nojoystick = 1;
      else if (strcmp ("-gravis", argv[i]) == 0)
         sw_gravis = 1;
      else if (strcmp ("-hr", argv[i]) == 0)
         sw_hr = 1;
      }

   //-- calibrate the joystick
   if (!sw_nojoystick)
      {        
      printf ("\nMake sure the joystick is centered...\n<hit any key>\n\n");
      getch();      
      if (initialise_joystick())
         sw_nojoystick = 0;
      else
         {
         if (sw_gravis)
            joy_type = JOY_TYPE_4BUTTON;
         else
            joy_type = JOY_TYPE_STANDARD;
         }
      }

   //-- initialize audio library
   if (!sw_nosound)
      {
      if (SetupAudio())
         {
         sw_nosound = 1;
         printf ("Error initializing audio.  Disabling sound.\n<Hit any key>\n");         	
         getch();
         }
      }

   printf("Loading Roms...\n");
   memset(RAM, 0, 0x20000);
   i = 0;
   while (cga->Roms[i].fileName)
      {
      if (!LoadRom(cga->Roms[i].loadAddr, cga->Roms[i].fileName, cga->Roms[i].romSize)) 
         {
         exit(1);
         }
      i++;
      }


   m6502Base = RAM;
   m6502reset();
   m6502MemoryRead = cga->MemRead;
   m6502MemoryWrite = cga->MemWrite;

   install_timer();
   install_keyboard();

   SetupVideo();

   //-- check for high refresh rate
   if (sw_hr)
      outportb(0x3c2, 0xef);
      
   DisplayForegroundChars();   

   // if 360 x 360 mode, display the backdrop - future feature
   if (sw_v360 && cga->GameID == VANGUARD)
      {
      BITMAP *bmp;
      PALLETE pal;
      get_pallete(pal);
      bmp = load_pcx("game.pcx", pal);
      if (bmp)
         {
         blit (bmp, screen, 0,0, 0,0, 360,360);
         blit (bmp, screen, 0,0, 0,360, 360,360);
         destroy_bitmap(bmp);
         }
      }

   //-- start audio
   if (!sw_nosound)
      {
      PlaySample(0, samples, 32, 1000, 0, 1);
      PlaySample(1, samples, 32, 1000, 0, 1);
      APlayVoice(hVoice[2], lpWave[2]);
      APlayVoice(hVoice[3], lpWave[3]);
      APlayVoice(hVoice[4], lpWave[4]);
      PlaySample(5, samples, 32, 1000, 0, 1);
      }

#ifdef SOUNDTEST
   fptr = fopen ("sound.txt", "w");
#endif

   memset (Background, 0, 0x8000);
   memset (Foreground, 0, 0x8000);

   // run 
   RAM[cga->DipSwitch] = sw_men;
   install_int(PlayAudio, 33);
   Exec();

#ifdef SOUNDTEST
   fclose (fptr);
#endif

   SaveHiScore();

   set_gfx_mode(GFX_TEXT,80,25,0,0);   

   //-- free audio memory
   if (!sw_nosound)
      {
      ShutdownAudio();
      }

   //-- destroy the bitmaps
   destroy_bitmap(bchars);
   destroy_bitmap(fchars);
   destroy_bitmap(piccie);
   allegro_exit();

#ifdef DUMPRW
   fptr = fopen ("write.txt", "w");
   for (i=0; i<0x2000; i++)
      {
      fprintf (fptr, "%04X: %5d,   %04X: %5d,   %04X: %5d,   %04X: %5d,   %04X: %5d\n",
         i, Background[i], i+0x2000, Background[i+0x2000], 
         i+0x4000, Background[i+0x4000], i+0x6000, Background[i+0x6000], 
         i+0x8000, Background[i+0x8000]);
      }
   fclose (fptr);
   fptr = fopen ("read.txt", "w");
   for (i=0; i<0x2000; i++)
      {
      fprintf (fptr, "%04X: %5d,   %04X: %5d,   %04X: %5d,   %04X: %5d,   %04X: %5d\n",
         i, Foreground[i], i+0x2000, Foreground[i+0x2000], 
         i+0x4000, Foreground[i+0x4000], i+0x6000, Foreground[i+0x6000], 
         i+0x8000, Foreground[i+0x8000]);
      }
   fclose (fptr);
#endif

   exit(0);
}

int LoadRom(int Address, char * Name, size_t Size) 
{                                     
   FILE * fp;
   char file[80];
   sprintf (file, "%s\\%s", cga->Path, Name);
   
   //-- open the ROM file
   printf("Loading %s ...", file);
   if ((fp = fopen(file, "rb")) == NULL)
   {
      printf("failed on open\n");
      return FALSE;
   }

   //-- read it into RAM image
   if (fread(RAM + Address, 1, Size, fp) != Size) {
         printf("failed on read\n");
         return FALSE;
   }
   fclose(fp);

   printf("\n");
   return TRUE;
}

//-- run emulation with masked interrupt every 12000 cycles
void Exec()
{
	BYTE rtn = INT_IRQ;
   while (rtn != INT_QUIT)
      {
      m6502exec(INTCOUNT);
      rtn = Interrupt();
      if (rtn == INT_NMI)
         {
         //-- force a non-masked interrupt
         m6502nmi();
         }
      else if (rtn == INT_IRQ)
         {
         //-- force a masked interrupt
         m6502int();
         }
      }
}

void hexdump(BYTE Data)
{
#ifdef SOUNDTEST
   fprintf (fptr, "b%01d", (Data&0x80)>>7);	
   fprintf (fptr, "%01d", (Data&0x40)>>6);	
   fprintf (fptr, "%01d", (Data&0x20)>>5);	
   fprintf (fptr, "%01d", (Data&0x10)>>4);	
   fprintf (fptr, "%01d", (Data&0x08)>>3);	
   fprintf (fptr, "%01d", (Data&0x04)>>2);	
   fprintf (fptr, "%01d", (Data&0x02)>>1);	
   fprintf (fptr, "%01d - %d\n", (Data&0x01), LoopCounter);	
#endif
}

void PlayAudio()
{
   if (pause)
      return;

   if (!sw_nosound) 
      {
      // output music
      if (NoSound == FALSE)
         {
         if (!NoSound0)
            AdjustSample (0, (32000 / (256-RAM[Sound0++])) * 16, 255);
         else
            AdjustSample (0, 1000, 0);
         if (!NoSound1)
            AdjustSample (1, (32000 / (256-RAM[Sound1++])) * 16, 255);
         else
            AdjustSample (1, 1000, 0);
         if (!NoSound2)
            AdjustSample (5, (32000 / (256-RAM[Sound2++])) * 16, 255);
         else
            AdjustSample (5, 1000, 0);
         }
      else
         {
         AdjustSample (0, 1000, 0);
         AdjustSample (1, 1000, 0);
         AdjustSample (5, 1000, 0);
         }
      AUpdateAudio();

      // check for overflow
      if (Sound0 > Sound0Base + 255)
         Sound0 = Sound0Base;
   
      if (Sound1 > Sound1Base + 255)
         Sound1 = Sound1Base;

      if (Sound2 > Sound2Base + 255)
         Sound2 = Sound2Base;
      }

   LoopCounter++; 

   if (sw_throttle && LoopCounter == 120)
      {
      int fps2 = intr_ticks >> 2;
      if (fps2 > 60)
         {
         int mpf = 10000 / fps2;
         throttle = (167 - mpf) / 10;
         }
      }
}

//-- out to sound port 0
int  VSoundOut0( WORD A,  BYTE B)
{
#ifdef SOUNDTEST
   if (sound_test)
      {
      fprintf (fptr, "%X: %02X - ", A, B);
      hexdump(B);
      }
#endif

   if (!sw_nosound)                                                
      {
      Sound0Base = ((B & 0x07) << 8) + cga->Sound1;
      if (Sound0Base != oSound0Base)
         {
         oSound0Base = Sound0Base;
         Sound0 = Sound0Base;
         }

      if (B & 0x20 && !(LastPort1 & 0x20))
         {
         APrimeVoice(hVoice[4], lpWave[4]);
         ASetVoiceVolume(hVoice[4], 64);
         AStartVoice(hVoice[4]);            
         }
      else if (!(B & 0x20) && LastPort1 & 0x20)
         AStopVoice(hVoice[4]);

      if (B & 0x40 && !(LastPort1 & 0x40))
         {
         APrimeVoice(hVoice[2], lpWave[2]);
         ASetVoiceVolume(hVoice[2], 64);
         AStartVoice(hVoice[2]);            
         }
      else if (!(B & 0x20) && LastPort1 & 0x20)
         AStopVoice(hVoice[4]);

      if (B & 0x80 && !(LastPort1 & 0x80))
         {
         APrimeVoice(hVoice[3], lpWave[3]);
         ASetVoiceVolume(hVoice[3], 64);
         AStartVoice(hVoice[3]);            
         }

      if (B & 0x10)
         {
         NoSound0 = FALSE;
         }
      else
         {
         }


      if (B & 0x08)
         {
         NoSound0 = TRUE;
         }
      }

   LastPort1 = B;
   RAM[cga->Port1] = B;
   return 0;
}

//-- out to sound port 0
int  FSoundOut0( WORD A,  BYTE B)
{
#ifdef SOUNDTEST
   if (sound_test)
      {
      fprintf (fptr, "%X: %02X - ", A, B);
      hexdump(B);
      }
#endif

   if (pause == TRUE)
      return 0;

   if (!sw_nosound)                                                
      {
      Sound0Base = ((B & 0x07) << 8) + cga->Sound1;
      if (Sound0Base != oSound0Base)
         {
         oSound0Base = Sound0Base;
         Sound0 = Sound0Base;
         }

      if (B & 0x20 && !(LastPort1 & 0x20))
         {
         APrimeVoice(hVoice[4], lpWave[4]);
         ASetVoiceVolume(hVoice[4], 64);
         AStartVoice(hVoice[4]);            
         }
      else if (!(B & 0x20) && LastPort1 & 0x20)
         AStopVoice(hVoice[4]);

      if (B & 0x40 && !(LastPort1 & 0x40))
         {
         APrimeVoice(hVoice[2], lpWave[2]);
         ASetVoiceVolume(hVoice[2], 64);
         AStartVoice(hVoice[2]);            
         }
      else if (!(B & 0x20) && LastPort1 & 0x20)
         AStopVoice(hVoice[4]);

      if (B & 0x80 && !(LastPort1 & 0x80))
         {
         APrimeVoice(hVoice[3], lpWave[3]);
         ASetVoiceVolume(hVoice[3], 64);
         AStartVoice(hVoice[3]);            
         }

      if (B & 0x08)
         {
         NoSound0 = FALSE;
         }
      else
         {
         NoSound0 = TRUE;
         Sound0 = Sound0Base;
         }

      if (B & 0x10)
         {
         NoSound2 = FALSE;
         }
      else
         {
         NoSound2 = TRUE;
         Sound2 = Sound2Base;
         }
      }

   LastPort1 = B;
   RAM[cga->Port1] = B;
   return 0;
}

//-- out to sound port 0
int  NSoundOut0( WORD A,  BYTE B)
{
#ifdef SOUNDTEST
   if (sound_test)
      {
      fprintf (fptr, "%X: %02X - ", A, B);
      hexdump(B);
      }
#endif

   if (!sw_nosound)                                                
      {
      Sound0Base = ((B & 0x07) << 8) + cga->Sound1;
      if (Sound0Base != oSound0Base)
         {
         oSound0Base = Sound0Base;
         Sound0 = Sound0Base;
         }

      if (B & 0x20 && !(LastPort1 & 0x20))
         {
         APrimeVoice(hVoice[4], lpWave[4]);
         ASetVoiceVolume(hVoice[4], 64);
         AStartVoice(hVoice[4]);            
         }
      else if (!(B & 0x20) && LastPort1 & 0x20)
         AStopVoice(hVoice[4]);

      if (B & 0x40 && !(LastPort1 & 0x40))
         {
         APrimeVoice(hVoice[2], lpWave[2]);
         ASetVoiceVolume(hVoice[2], 64);
         AStartVoice(hVoice[2]);            
         }
      else if (!(B & 0x20) && LastPort1 & 0x20)
         AStopVoice(hVoice[4]);

      if (B & 0x80 && !(LastPort1 & 0x80))
         {
         APrimeVoice(hVoice[3], lpWave[3]);
         ASetVoiceVolume(hVoice[3], 64);
         AStartVoice(hVoice[3]);            
         }

      if (B & 0x08 && cga->Sound1)
         {
         NoSound0 = FALSE;
         }
      else
         {
         NoSound0 = TRUE;
         }

      if (B & 0x10)
         {
         }
      }

   LastPort1 = B;
   RAM[cga->Port1] = B;
   return 0;
}

//-- out to sound port 1
int  SoundOut1( WORD A,  BYTE B)
{
#ifdef SOUNDTEST
   if (sound_test)
      {
      fprintf (fptr, "%X: %02X - ", A, B);
      hexdump(B);
      }
#endif

   if (!sw_nosound)                                                
      {
      Sound1Base = ((B & 0x07) << 8) + cga->Sound2;
      if (Sound1Base != oSound1Base)
         {
         oSound1Base = Sound1Base;
         Sound1 = Sound1Base;
         }

      if (B & 0x08)
         {
         NoSound1 = FALSE;
         }
      else
         {
         NoSound1 = TRUE;
         }
      }

   LastPort2 = B;
   RAM[cga->Port2] = B;
   return 0;
}

//-- out to sound port 2 ??
int  SoundOut2( WORD A,  BYTE B)
{
#ifdef SOUNDTEST
   if (sound_test)
      {
      fprintf (fptr, "%X: %02X - ", A, B);
      hexdump(B);
      }
#endif

   if (!sw_nosound)
      {
      }

  	LastPort3 = B;
   RAM[cga->Port3] = B;
   return 0;
}

//-- out to sound port 3 ??
int  ColorOut( WORD A,  BYTE B)
{
#ifdef SOUNDTEST
   if (sound_test)
      {
      fprintf (fptr, "%X: %02X - ", A, B);
      hexdump(B);
      }
#endif

   if (!sw_nosound && cga->GameID == FANTASY)
      {
      Sound2Base = (swapsound[(B & 0x70) >> 4] << 8) + cga->Sound3;
      if (Sound2Base != oSound2Base)
         {
         oSound2Base = Sound2Base;
         Sound2 = Sound2Base;
         }
      }

 	LastPort4 = B;
   RAM[cga->Port4] = B;
   return 0;
}

//-- out to speech port ??
int  SpeechOut( WORD A,  BYTE B)
{
#ifdef SOUNDTEST
   if (sound_test)
      {
//      fprintf (fptr, "%X: %02X - ", A, B);
//      hexdump(B);
      }
#endif

   if (!sw_nosound)
      {
	   //-- out to speech port
      }

   RAM[cga->Speech] = B;
   return 0;
}

WORD ForegroundUpdate( WORD A)
{
   Foreground[A]++;
   return (WORD)(RAM[A]);
}

int BackgroundUpdate( WORD A,  BYTE B)
{
   Background[A]++;
   RAM[A] = B;
   return 0;
}

//-- attempting to read coins, start port
WORD ReadCoin( WORD A)
{
   BYTE valCoin = 0;

   if (key[KEY_1])
      {
      //-- start 1 player game
      valCoin |= 0x80;
      if (!score_loaded)
         {
         score_loaded = 1;
         }
      }
   else if (key[KEY_2])
      {
      //-- start 2 player game
      valCoin |= 0x40;
      if (!score_loaded)
         {
         score_loaded = 1;
         }
      }  
      	
   return (WORD)valCoin;
}

//-- attempting to read player 1 controls
WORD ReadPlayer1( WORD A)
{
   BYTE valP1 = 0;
   if (key[KEY_UP])
      valP1 |= 0x20;  // player 1 move up
      
   if (key[KEY_DOWN])
      valP1 |= 0x10;  // player 1 move down
      
   if (key[KEY_LEFT])
      valP1 |= 0x80;  // player 1 move left
      
   if (key[KEY_RIGHT])
      valP1 |= 0x40;  // player 1 move right
      
   if (key[KEY_W])
      valP1 |= 0x02;  // player 1 shoot up
      
   if ((sw_useX && key[KEY_X]) || (!sw_useX && key[KEY_S]))
      valP1 |= 0x01;  // player 1 shoot down
      
   if (key[KEY_A])
      valP1 |= 0x08;  // player 1 shoot left

   if (key[KEY_D])
      valP1 |= 0x04;  // player 1 shoot right

   if (!sw_nojoystick)
      {
      if (joy_up)
         valP1 |= 0x20;  // player 1 move up
      
      if (joy_down)
         valP1 |= 0x10;  // player 1 move down
      
      if (joy_left)
         valP1 |= 0x80;  // player 1 move left
      
      if (joy_right)
         valP1 |= 0x40;  // player 1 move right

      if (!sw_gravis && joy_next_shot)
         {
         valP1 |= joy_next_shot;
         joy_next_shot = 0;
         }
      else
         {
         if (joy_b1)
            {
            valP1 |= 0x08;  // player 1 shoot left
            if (!sw_gravis)
               joy_next_shot = 0x04;   // shoot right next
            }
         if (joy_b2)
            {
            if (!sw_gravis)
               {
               valP1 |= 0x02;  // player 1 shoot up and down
               joy_next_shot = 0x01;
               }
            else
               valP1 |= 0x02;  // player 1 shoot up
            }
         if (sw_gravis && joy_b3)
            valP1 |= 0x01;  // player 1 shoot down
         if (sw_gravis && joy_b4)
            valP1 |= 0x04;  // player 1 shoot right
         }
      }      

   return (WORD)valP1;
}
              
//-- attempting to read player 2 controls
WORD ReadPlayer2( WORD A)
{
   BYTE valP2 = 0;
   if (key[KEY_UP])
      valP2 |= 0x20;  // player 2 move up
      
   if (key[KEY_DOWN])
      valP2 |= 0x10;  // player 2 move down
      
   if (key[KEY_LEFT])
      valP2 |= 0x80;  // player 2 move left
      
   if (key[KEY_RIGHT])
      valP2 |= 0x40;  // player 2 move right
      
   if (key[KEY_W])
      valP2 |= 0x02;  // player 2 shoot up
      
   if ((sw_useX && key[KEY_X]) || (!sw_useX && key[KEY_S]))
      valP2 |= 0x01;  // player 2 shoot down
      
   if (key[KEY_A])
      valP2 |= 0x08;  // player 2 shoot left
      
   if (key[KEY_D])
      valP2 |= 0x04;  // player 2 shoot right

   if (!sw_nojoystick)
      {
      if (joy_up)
         valP2 |= 0x20;  // player 2 move up
      
      if (joy_down)
         valP2 |= 0x10;  // player 2 move down
      
      if (joy_left)
         valP2 |= 0x80;  // player 2 move left
      
      if (joy_right)
         valP2 |= 0x40;  // player 2 move right
      
      if (!sw_gravis && joy_next_shot)
         {
         valP2 |= joy_next_shot;
         joy_next_shot = 0;
         }
      else
         {
         if (joy_b1)
            {
            valP2 |= 0x08;  // player 2 shoot left 
            if (!sw_gravis)
               joy_next_shot = 0x04;   // next shot right
            }
         
         if (joy_b2)
            {
            if (!sw_gravis)
               {
               valP2 |= 0x02;  // player 2 shoot up and down
               joy_next_shot = 0x01;
               }
            else
               valP2 |= 0x02;  // player 2 shoot up
            }
         if (sw_gravis && joy_b3)
            valP2 |= 0x01;  // player 2 shoot down
         if (sw_gravis && joy_b4)
            valP2 |= 0x04;  // player 2 shoot right
         }
      }      

   return (WORD)valP2;
}

void DumpRom(char *nm, BYTE *bp, WORD len)
{
   FILE *f5;
   f5 = fopen (nm, "wb");
   fwrite (bp, len, 1, f5);
   fclose (f5);
}

void OutRom(char *nm, BYTE *bp, WORD len)
{
   FILE *f5;
   int i;
   f5 = fopen (nm, "w");
   for (i=0; i<len; i++)
	fprintf (f5, "%04X: %d\n", i, bp[i]);
   fclose (f5);
}

void DisplayGraphics()
{
   char *start, *ptr;
   unsigned char c;
   int i, j; 
   
   // display characters, accounting for rotation
   start = RAM + cga->BackgroundRAM + 31 * 32;
   for (i=0; i<32; i++)
      {
   	ptr = start;  
      for (j=0; j<32; j++)
         {
         c = *ptr;
         if (c != 48)
            {
            DrawCharacter (c, j<<3, i<<3, (BYTE)*(ptr+(cga->ColorRAM-cga->BackgroundRAM)));
            }
   	   ptr -= 32;
  	      }
  	   start++;
     	}

   //-- display the forground characters as sprites
   start = RAM + cga->ForegroundRAM + 31 * 32;
   for (i=0; i<32; i++)
      {
     	ptr = start;  
      for (j=0; j<32; j++)
         {
         c = *ptr;
         if (c != 48)
 		      {
            //-- updated sprites are read from 0x1000-0x1fff
            //-- sprite color palette is from 0xc000-0xcfff
            DrawSprite (c, j<<3, i<<3, (BYTE)*(ptr+(cga->ColorRAM-cga->ForegroundRAM))); 
 		      }

         ptr -= 32;
   	   }
	   start++;
   	}
}

void WarlordsDisplayGraphics()
{
   char *start, *ptr;
   unsigned char c;
   int i, j; 
   
   // display characters, accounting for rotation
   start = RAM + cga->BackgroundRAM;
   for (i=0; i<32; i++)
      {
      for (j=0; j<32; j++)
         {
         c = *(start+i*32+j);
         WarlordsDrawCharacter (c, j<<3, i<<3, (BYTE)*(ptr+(cga->ColorRAM-cga->BackgroundRAM)));
  	      }
     	}
}

void CCastlesDisplayGraphics()
{
   char *start;
   unsigned char c;
   int i, j; 
   
   // display characters, accounting for rotation
   start = RAM + cga->BackgroundRAM;
   for (i=0; i<32; i++)
      {
      for (j=0; j<32; j++)
         {
         c = *(start+i*32+j);
         CCastlesDrawCharacter (c, j<<3, i<<3);
  	      }
     	}
}

//-- masked interrupt --- most of the emulation takes place here
BYTE Interrupt(void)
{
   int rtn = INT_IRQ;

   intr_ticks++;
   if (sw_skip && (intr_ticks & 1))
      return rtn;

   //-- always clear the background
   clear_to_color(piccie, cga->Color[LastPort4 & 7]);

   cga->Display();

   //-- check for user exit
   if (key[KEY_ESC])
      rtn = INT_QUIT;

   //-- check for pause
   if (key[KEY_P])
      {
      pause = TRUE;

      //-- TODO: if sound, need to turn volume off
      if (!sw_nosound)
         {
         remove_int(PlayAudio);
         AdjustSample (0, 1000, 0);
         AdjustSample (1, 1000, 0);
         AdjustSample (5, 1000, 0);
         AUpdateAudio();
         AStopVoice(hVoice[2]);
         AStopVoice(hVoice[3]);
         AStopVoice(hVoice[4]);
         }

      //-- display a pause message
      PrintStr("HIT P TO CONTINUE", 128, 0x50);
      }
   //-- check for coin 1 - requires nmi
   else if (key[KEY_3])
      {
      RAM[cga->Coin] |= 0x01;
      clear_keybuf();
      rtn = INT_NMI;
      }
   //-- check for coin 2 - requires nmi
   else if (key[KEY_4])
      {
      RAM[cga->Coin] |= 0x02;
      clear_keybuf();
      rtn = INT_NMI;
      }   
   //-- display frames per second
   else if (key[KEY_F1])
      {
      ShowFPS ^= 1;
      rest(20);
      }
   //-- screen dump
   else if (key[KEY_F2])
      {
      BITMAP *bmp;
      PALLETE pal;
      char file[80];
      sprintf (file, "%s\\game.pcx", cga->Path);
      get_pallete(pal);
      bmp = create_sub_bitmap(screen, 0,0, 224,256);
      save_pcx(file, bmp, pal);
      destroy_bitmap(bmp);
      }
   else if (key[KEY_F3])
      {
      SaveGame();
      }
   else if (key[KEY_F4])
      {
      LoadGame();
      }
   else if (key[KEY_F5])
      {
      OutRom ("rom.b", RAM, 0x400);
      }
   else if (key[KEY_F6])
      {
      OutRom ("rom.a", RAM, 0x400);
      }
   //-- sound test
   else if (key[KEY_F12])
      {
      sound_test ^= 1;
      rest(20);
      }
   //-- check the joystick
   if (!sw_nojoystick)
      {
      poll_joystick();	
      }

   if (cga->GameID == CCASTLES)
      {
      char buf[20];
      sprintf (buf, "%3d  %3d", RAM[0], RAM[1]);
      textout (piccie, font, buf, 32, 0, 0xff);
      }

#ifdef DUMPDATA
   PrintVal (0x2100, 248, 0x20);
   PrintVal (0x2101, 248, 0x40);
   PrintVal (0x2102, 248, 0x60);
   PrintVal (0x2103, 248, 0x80);
#endif

   if (ShowFPS && LoopCounter >= 30)
      {
      fps = intr_ticks / (LoopCounter / 30);
      PrintNum (fps, 248, 0x20);
      }

#ifdef SOUNDTEST
   if (sound_test)
      PrintStr("ON", 248, 0x20);
   else
      PrintStr("OFF", 248, 0x20);
#endif

   if (sw_throttle && throttle)
      rest(throttle);

   // blast the video work buffer to screen
   if (sw_noxblit)
      {
      if (sw_v360)
         blit(piccie, screen, 32,0, 68,84, 224,256);
      else
         blit(piccie, screen, 32,0, 0,0, 224,256);
      }
   else if (sw_xvga)
      {
      xblast (piccie, 224, 32);
      }
   else      
      {  
      if (sw_v360)
         {
         if (ActivePage)
            xb(piccie, screen, 32,0, 68,84, 224,256, 0);
         else
            xb(piccie, screen, 32,0, 68,444, 224,256, 360);
         ActivePage ^= 1;
         }
      else
         {
         if (ActivePage)
            xb(piccie, screen, 32,0, 0,0, 224,256, 0);
         else
            xb(piccie, screen, 32,0, 0,256, 224,256, 256);
         ActivePage ^= 1;
         }
      }

   if (pause)
      {
      clear_keybuf();
      while (!key[KEY_P])
         continue;
      clear_keybuf();

      //-- TODO: if sound, turn on
      if (!sw_nosound)
         {
         install_int(PlayAudio, 33);
         AStartVoice(hVoice[2]);
         AStartVoice(hVoice[3]);
         AStartVoice(hVoice[4]);
         }

      pause = FALSE;
      }

   return rtn;
}

void CCastlesDrawCharacter (int c, int col, int row)
{
	int x, y;
   BYTE result = 0;
   BYTE *ROM1, *ROM2;
   BYTE b1, b2;

   ROM1 = RAM + cga->BChar1 + 16 * c;
   ROM2 = RAM + cga->BChar2 + 16 * c;

   for (y=0; y<8; y++)
      {
      for (x=0; x<2; x++)
         {
         b1 = *ROM1++;
         b2 = *ROM2++;
         result = (b1&0x80?8:0) + (b1&0x08?4:0) + (b2&0x80?2:0) + (b2&0x08?1:0);
         piccie->line[row+y][col+x*4] = result << 4;
         result = (b1&0x40?8:0) + (b1&0x04?4:0) + (b2&0x40?2:0) + (b2&0x04?1:0);
         piccie->line[row+y][col+x*4+1] = result << 4;
         result = (b1&0x20?8:0) + (b1&0x02?4:0) + (b2&0x20?2:0) + (b2&0x02?1:0);
         piccie->line[row+y][col+x*4+2] = result << 4;
         result = (b1&0x10?8:0) + (b1&0x01?4:0) + (b2&0x10?2:0) + (b2&0x01?1:0);
         piccie->line[row+y][col+x*4+3] = result << 4; 
         }
      }
}

//-- draw rotated character at specified position
void WarlordsDrawCharacter (int c, int col, int row, BYTE color)
{
	int x, y;
   int sr, sc;
   int sb1 = 0;
   int sb2 = 0;
   BYTE result = 0;
   BYTE res1, res2;
   BYTE mask;
   int hpos;
   int vpos;
   int ROM1, ROM2;

   //-- background character color palette selected using bits 4,5,6
   BYTE cPal = (color & 0x38) >> 3;

   ROM1 = cga->BChar1;
   ROM2 = cga->BChar2;

   // get the offset to the correct bit for each bitplane
   sb1 = c * 64;

   // for each bit in character
	sr  = 0;
   for (y = 0; y < 8; y ++)
      {
      // get a BYTE at a time
	   sc = 7;
      res1 = RAM[(sb1>>3)+ROM1];
      res2 = RAM[(sb1>>3)+ROM2];

      vpos = sr+row;

      // for each bit in BYTE
      for (x=0; x<8; x++)
         {
         hpos = sc+col;

         // find the color by adding the bitplane values
         mask = 0x80 >> x; 
         if ((res1 & mask) == mask) result = 1;
         if ((res2 & mask) == mask) result ^= 2;
         if (result)
            {
            if (sw_noxblit || sw_xvga)
               piccie->line[vpos][hpos] = cga->Color[(cPal<<2)+result];
            else
               // the custom blit expects to find data packed by plane per line
               piccie->line[vpos][(hpos>>2)+((hpos%4)<<6)] = cga->Color[(cPal<<2)+result];
            }
         sc--;
         sb1++;
         sb2++;
         result = 0;
         }

	   sr++;
      }
}

//-- draw rotated character at specified position
void DrawCharacter (int c, int col, int row, BYTE color)
{
	int x, y;
   int sr, sc;
   int sb1 = 0;
   int sb2 = 0;
   BYTE result = 0;
   BYTE res1, res2;
   BYTE mask;
   int hpos;
   int vpos;
   int ROM1, ROM2;

   //-- background character color palette selected using bits 4,5,6
   BYTE cPal = (color & 0x38) >> 3;

   ROM1 = cga->BChar1;
   ROM2 = cga->BChar2;

   if (cga->GameID == FANTASY)
      {
      ROM1 = (RAM[0x4D]&0x08 ? cga->BChar1 - 0x800 : cga->BChar1);
      ROM2 = (RAM[0x4D]&0x08 ? cga->BChar2 - 0x800 : cga->BChar2);
      }

   // get the offset to the correct bit for each bitplane
   sb1 = c * 64;

   // for each bit in character
	sc  = 7;
   for (y = 0; y < 8; y ++)
      {
      // get a BYTE at a time
	   sr = 0;
      res1 = RAM[(sb1>>3)+ROM1];
      res2 = RAM[(sb1>>3)+ROM2];

      //-- adjust horizontal scrolling
      hpos = sc+col+RAM[cga->HScroll];
      if (hpos >= 256) hpos -= 256;
      else if (hpos < 0) hpos += 256;

      // for each bit in BYTE
      for (x=0; x<8; x++)
         {
         //-- adjust vertical scrolling
         vpos = sr+row-RAM[cga->VScroll];
         if (vpos >= 256) vpos -= 256;
         else if (vpos < 0) vpos += 256;

         // find the color by adding the bitplane values
         mask = 0x80 >> x; 
         if ((res1 & mask) == mask) result = 1;
         if ((res2 & mask) == mask) result ^= 2;
         if (result)
            {
            if (sw_noxblit || sw_xvga)
               piccie->line[vpos][hpos] = cga->Color[(cPal<<2)+result];
            else
               // the custom blit expects to find data packed by plane per line
               piccie->line[vpos][(hpos>>2)+((hpos%4)<<6)] = cga->Color[(cPal<<2)+result];
            }
         sr++;
         sb1++;
         sb2++;
         result = 0;
         }

	   sc--;
      }
}

//-- draw rotated sprite at specified position
void DrawSprite (int c, int col, int row, BYTE color)
{
	int x, y;
   int sr, sc;
   int sb1 = 0;
   int sb2 = 0;
   BYTE result = 0;
   BYTE res1, res2;
   BYTE mask;

   //-- sprite color palette selected using bits 0,1,2
   BYTE cPal = color & 0x07;

   // get the offset to the correct bit for each bitplane
   sb1 = c * 64;

   // for each bit in sprite
	sc  = 7;
   for (y = 0; y < 8; y ++)
      {
      // get a BYTE at a time
	   sr = 0;
      res1 = RAM[(sb1>>3)+0x1000];
      res2 = RAM[(sb1>>3)+0x1800];

      // for each bit in BYTE
      for (x=0; x<8; x++)
         {
         // find the color by adding the bitplane values
         mask = 0x80 >> x; 
         if ((res1 & mask) == mask) result = 1;
         if ((res2 & mask) == mask) result ^= 2;
         if (result)
            {
            if (sw_noxblit || sw_xvga)
               piccie->line[sr+row][sc+col] = cga->Color[32+(cPal<<2)+result];
            else
               // the custom blit expects to find data packed by plane per line
               piccie->line[sr+row][((sc+col)>>2)+(((sc+col)%4)<<6)] = cga->Color[32+(cPal<<2)+result];
            }
         sr++;
         sb1++;
         sb2++;
         result = 0;
         }

	   sc--;
      }
}

//-- play a sample
void PlaySample (int channel, BYTE *data, int len, int freq, int vol, int loop)
{
   //-- sanity check
	if (channel >= NUMVOICES)
      return;

   memcpy(lpWave[channel]->lpData, data, len);
   lpWave[channel]->wFormat = AUDIO_FORMAT_8BITS | AUDIO_FORMAT_MONO;
   if (loop)
      lpWave[channel]->wFormat |= AUDIO_FORMAT_LOOP;
   lpWave[channel]->dwLoopStart = 0;
   lpWave[channel]->dwLoopEnd = len;
   lpWave[channel]->dwLength = len;

   AWriteAudioData(lpWave[channel], 0, len);
   APlayVoice(hVoice[channel], lpWave[channel]);
   ASetVoiceFrequency(hVoice[channel], freq);
   ASetVoiceVolume(hVoice[channel], vol>>2);
}

void AdjustSample (int channel, int freq, int vol)
{
   //-- sanity check
	if (channel >= NUMVOICES)
      return;

   if (freq == 512000) vol = 0;
   ASetVoiceFrequency(hVoice[channel], freq);
   ASetVoiceVolume(hVoice[channel], vol>>2);
}

// capture and prerotate all the background characters to a bitmap
void DisplayBackgroundChars(void)
{
   int x,y,cx,cy;
   int sr, sc;
   int sb1 = 0;
   int sb2 = 0;

   BYTE result = 0;

   for (cy = 0; cy < 8; cy ++)
   {
      for (cx = 0; cx < 32; cx ++)
      {
	   sc  = 7;
         for (y = 0; y < 8; y ++)
         {
	      sr = 0;
            for (x = 0; x < 8; x ++)
            {
               if ( GetBit(sb1, cga->BChar1) )
                 result = 1; 
               if ( GetBit(sb2, cga->BChar2) )
                 result ^= 2;

               //-- get the correct palette color
               bchars->line[cy*8 + sr][cx*8 + sc] = result;

               result = 0;
               sb1++;
               sb2++;
		         sr++;
            }
	   sc--;
         }
      }
   }
}

// capture and prerotate all the foreground characters to a bitmap
void DisplayForegroundChars(void)
{
   int x,y,cx,cy;
   int sr, sc;
   int sb1 = 0;
   int sb2 = 16384;

   BYTE result = 0;
 
   for (cy = 0; cy < 8; cy ++)
   {
      for (cx = 0; cx < 32; cx ++)
      {
	      sc  = 7;
         for (y = 0; y < 8; y ++)
         {
	         sr = 0;
            for (x = 0; x < 8; x ++)
            {
               if ( GetBit(sb1, 0x8000) )
                 result = 1;
               if ( GetBit(sb2, 0x8000) )
                 result ^= 2;
               
               fchars->line[cy*8 + sr][cx*8 + sc] = result; 

               result = 0;
               sb1++;
               sb2++;
		         sr++;
            }
	      sc--;            
         }
      }
   }

   DisplayBackgroundChars();

   // display them
   if (sw_dumpgraphics)
      {
      blit(fchars, screen, 0,0, 0,72, 256,64);
      blit(bchars, screen, 0,0, 0,0, 256,64);
      readkey();
      }
}


// return specific bit from memory location
WORD GetBit(int  startbit, unsigned int off)
{
   WORD nBYTE = startbit / 8 + off;
   WORD nbit = startbit%8;
          
   switch (nbit)
   {
      case 0:
         return ((RAM[nBYTE] & 0x80) == 0x80);
         break;

      case 1:
         return ((RAM[nBYTE] & 0x40) == 0x40);
         break;

      case 2:
         return ((RAM[nBYTE] & 0x20) == 0x20);
         break;

      case 3:
         return ((RAM[nBYTE] & 0x10) == 0x10);
         break;

      case 4:
         return ((RAM[nBYTE] & 0x08) == 0x08);
         break;

      case 5:
         return ((RAM[nBYTE] & 0x04) == 0x04);
         break;

      case 6:
         return ((RAM[nBYTE] & 0x02) == 0x02);
         break;

      case 7:
         return ((RAM[nBYTE] & 0x01) == 0x01);
         break;

   }

   return 0;
}

// blit data by plane per line
void cBlit (int x, int y, int col, int row, int w, int h)
{
   int i, j;
   for (i=0; i<h; i++)
      for (j=0; j<w; j++)
         piccie->line[row+i][((col+j)>>2)+(((col+j)%4)<<6)] = fchars->line[y+i][x+j];
}

// display a value from a memory location
void PrintVal(int add, int row, int col)
{
   char buf[20];
   char *ptr;
   int val;

   val = RAM[add];
   sprintf (buf, "%u", val);
   ptr = buf;

   while (*ptr)
      {
      if (sw_noxblit || sw_xvga)
      	blit (fchars, piccie, (*ptr-'0')*8,0, col,row, 8,8);
      else
      	cBlit ((*ptr-'0')*8,0, col,row, 8,8);
      col += 8;
      ptr++;
   	}
}

// display a number
void PrintNum(int num, int row, int col)
{
   char buf[20];
   char *ptr;

   sprintf (buf, "%u", num);
   ptr = buf;

   while (*ptr)
      {
      if (sw_noxblit || sw_xvga)
   	   blit (fchars, piccie, (*ptr-'0')*8,0, col,row, 8,8);
      else
	      cBlit ((*ptr-'0')*8,0, col,row, 8,8);
      col += 8;
      ptr++;
   	}
}

// display a string
void PrintStr(char *str, int row, int col)
{
   int x, y;
   BYTE b;
   while (*str)
      {
      b = *str;
      if (b >= '0' && b <= '9')
         {
         y = 0;
         x = (b-'0')*8;
         }
      else if ((b >= 'A' && b <= 'Z') || (b >= 'a' && b <= 'z'))
         {
         b &= 0xdf;
         y = 0;
         x = (b-'A'+10)*8;
         }
      else if (b == '.')
         {
         y = 8;
         x = 48;
         }
      else
         {
         y = 8;
         x = 128;
         }
      if (sw_noxblit || sw_xvga)
      	blit (fchars, piccie, x, y, col,row, 8,8);
      else
      	cBlit (x, y, col,row, 8,8);
      col += 8;
      str++;
	   }
}







